// c_mfcc.cpp : Defines the entry point for the console application.
//

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>

#include <fstream>
#include <string>
#include <time.h>

#include "bass.h"
#include "tags.h"
#include "libmfcc.h"

#define NUM_COEFF 13
#define USE_MT

/* ! \brief Calculates the fingerprints for given audio files
 *
 * For audiofiles, specified by the first input file, this program at first calculates 13 MFCCs and then the mean and the variance over all windows (of the audio file) for all coefficients.
 * It also gathers tag information from the audio files and puts them into a list, containing the path to the file, the title, artist, album and year
 * 
 * \param argv[1]: the file containing the list of audio files
 * \param argv[2]: the output file in which the tag information is stored
 * \param argv[3]: the fingerprint output file. this file is in binary format
 */
int main(int argc, char* argv[])
{
	if (HIWORD(BASS_GetVersion())!=BASSVERSION) {
		printf("An incorrect version of BASS was loaded");
		return -1;
	}

	if(argc<=3) {
		printf("You need to specify an initial file list, an output list file and the output fingerprint file");
		return -1;
	}

	BASS_CHANNELINFO info;

	BASS_SetConfig(BASS_CONFIG_UPDATEPERIOD,0); //nothing to play
	BASS_Init(0,44100,0,0,NULL); //no output device

	DWORD chan;
	DWORD pos;

	double numReads;
	double mean[NUM_COEFF];
	double var[NUM_COEFF];

	double dBytes;
	double dProc;
	clock_t tStart;
	clock_t tEnd;
	clock_t tSum = 0;

	FILE* fFingerprint;
    fFingerprint = fopen(argv[3], "a+b");

	std::ofstream fileList(argv[2], std::ios_base::app);

	std::ifstream input( argv[1] );
	for( std::string line; std::getline( input, line ); )
	{
		tStart = clock();

		numReads = 0.0;
		for(int c = 0; c < NUM_COEFF; c++) {mean[c]=0.0; var[c]=0.0;}

		std::size_t found = line.find(";");
		std::string fileName = line.substr(0,found);

		printf("%s processing:\n",fileName.c_str());

		//open the file
		if ((chan=BASS_StreamCreateFile(FALSE,fileName.c_str(),0,0,BASS_STREAM_DECODE)) || (chan=BASS_StreamCreateURL(fileName.c_str(),0,BASS_STREAM_DECODE|BASS_STREAM_BLOCK,0,0))) {
			pos=BASS_ChannelGetLength(chan,BASS_POS_BYTE);

			if (pos==-1) // length not available
				printf(" * streaming file\n");
			else
				printf(" * streaming file [%I64u bytes]\n",pos);
		}
		else {
			printf(" * could not read mp3 file\n");
			continue;
		}
		dBytes = (double) pos;
		dProc = 0.0;


		//save info
		const char* t = TAGS_Read( chan, "%TITL;%ARTI;%ALBM;%YEAR" );
		fileList << fileName.c_str() << ";" << t << std::endl;
		printf(" * info saved\n");


		//calc the mfcc
		while (!_kbhit() && BASS_ChannelIsActive(chan)) {
			float fft[1024]; //real fft values
			double fftD[1024];

			//we need to get the complex data out, cause otherwise only the first half of it is returned
			BASS_ChannelGetData(chan,fft,BASS_DATA_FFT2048); //mono fft is taken care of in this function
			for(int i=0; i<1024; i++) fftD[i] = (double)fft[i]; //conversion

			for(unsigned int c = 0; c < NUM_COEFF; c++) //calc the mel freq spectrum for each coefficient
			{
				double mfcc_result = GetCoefficient(fftD, 44100, 48, 1024, c);

				mean[c] += mfcc_result;
				var[c] += mfcc_result * mfcc_result;
			}

			numReads += 1.0f;

			dProc = (double) BASS_ChannelGetPosition(chan,BASS_POS_BYTE);
			printf(" * %3.1lf %%\r",(dProc/dBytes)*100.0);
		}
		
		for(int c = 0; c < NUM_COEFF; c++) {
			mean[c] /= numReads; 
			var[c] /= numReads;
			var[c] -= mean[c]*mean[c];
		}
		fwrite(mean, sizeof(double), NUM_COEFF, fFingerprint);
		fwrite(var, sizeof(double), NUM_COEFF, fFingerprint);

		BASS_StreamFree(chan);


		tEnd = clock() - tStart;
		printf(" * fingerprint saved in %.3f seconds\n",(float)tEnd/CLOCKS_PER_SEC);
		tSum += tEnd;
	}

	input.close();
	fclose(fFingerprint);
	fileList.close();

	BASS_Free();

	printf("DONE in %.3f seconds!",(float)tSum/CLOCKS_PER_SEC);
	getchar();
	return 0;
}

